/***************************************
  $Header: /home/amb/wwwoffle/RCS/wwwoffles.c 1.3 1997/01/17 20:46:17 amb Exp $

  WWWOFFLE - World Wide Web Offline Explorer - Version 0.9a.
  A server to fetch the required pages.
  ******************/ /******************
  Written by Andrew M. Bishop

  This file Copyright 1996,97 Andrew M. Bishop
  It may be distributed under the GNU Public License, version 2, or
  any higher version.  See section COPYING of the GNU Public license
  for conditions under which this file may be redistributed.
  ***************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <unistd.h>
#include <fcntl.h>

#include "wwwoffle.h"
#include "sockets.h"
#include "errors.h"


static void usage(void);

typedef enum _Mode
{
 None,                          /*+ Undecided. +*/
 Real,                          /*+ From server host to cache and client. +*/
 RealNoCache,                   /*+ From server host to client. +*/
 Fetch,                         /*+ From server host to cache. +*/
 Spool,                         /*+ From cache to client. +*/
 SpoolGet,                      /*+ Not in cache so record request in outgoing. +*/
 Get,                           /*+ Record request in outgoing. +*/
 Index                          /*+ An index. +*/
}
Mode;

/*+ The mode of the fetch +*/
static Mode mode=None;

/*+ The read (and write) file descriptor for the client. +*/
static int client=-1;

/*+ The file descriptor to read and write to the remote host. +*/
static int server=-1;

/*+ The read or write file descriptor for the spool file. +*/
static int spool=-1;

/*+ The read or write file descriptor for the outgoing spool directory. +*/
static int outgoing=-1;


/*++++++++++++++++++++++++++++++++++++++
  The main program.
  ++++++++++++++++++++++++++++++++++++++*/

int main(int argc, char** argv)
{
 char *buffer,*url;
 int n;
 FILE *client_file=NULL,*outgoing_file=NULL;
 char *host,*path,*args;
 int port=80;

 /* Parse the command line options */

 if(argc!=3 && argc!=4)
    usage();

 if(!strcmp(argv[1],"-real"))
    mode=Real;
 else if(!strcmp(argv[1],"-fetch"))
    mode=Fetch;
 else if(!strcmp(argv[1],"-spool"))
    mode=Spool;
 else if(!strcmp(argv[1],"-get"))
    mode=Get;
 else
    exit(1);

 /* Initialise things. */

 InitErrorHandler(argv[0]);

 signal(SIGPIPE,SIG_IGN);

 /* Set up the file handle to read the request from. */

 client=atoi(argv[2]);

 if(client<=0)
    PrintMessage(Fatal,"Cannot use client file descriptor %d.",client);

 client_file=fdopen(client,"r");
 if(!client_file)
    PrintMessage(Fatal,"Cannot use fdopen on the client file descriptor %d.",client);

 if(mode==Fetch)
   {
    outgoing=OpenOutgoingSpoolFile(1);

    if(outgoing==-1)
       PrintMessage(Fatal,"Cannot open the spooled outgoing request to read.");
    else if(outgoing==-2)
      {PrintMessage(Inform,"No more spooled outgoing requests.");return(3);}

    outgoing_file=fdopen(outgoing,"r");
    if(!outgoing_file)
       PrintMessage(Fatal,"Cannot use fdopen on the outgoing file descriptor %d.",outgoing);
   }

 /* Get the URL from the request. */

 if(mode==Real || mode==Spool || mode==Get)
    buffer=GetURL(client_file,&url);
 else /* mode==Fetch */
    buffer=GetURL(outgoing_file,&url);

 if(!url)
   {
    if(mode!=Get)
       ServerErrorMessage(client,buffer?"Cannot parse the HTTP request":"The HTTP request was empty");

    PrintMessage(Fatal,"Could not parse HTTP request (%s).",buffer?"Parse error":"Empty request");
   }

 PrintMessage(Inform,"URL = %s",url);

 SplitURL(url,&host,&path,&args);

 /* Change the mode as required. */

 if((mode==Real || mode==Spool) && !*host)
   {
    if(!strncmp(url,"/outgoing/",10))
      {
       char *copy,*u;

       copy=(char*)malloc(strlen(buffer)+1);
       u=strstr(buffer,url); *u=0;
       strcpy(copy,buffer);
       strcat(copy,"http://");
       strcat(copy,&url[10]);
       strcat(copy,&u[strlen(url)]);
       free(buffer);
       buffer=copy;

       copy=(char*)malloc(strlen(url)+1);
       strcpy(copy,"http://");
       strcat(copy,&url[10]);
       free(url);
       url=copy;

       SplitURL(url,&host,&path,&args);

       if(mode==Spool)
          mode=SpoolGet;
      }
    else
       mode=Index;
   }
 else if(mode==Spool && !strncmp(host,"localhost",9))
    mode=RealNoCache;
 else if(mode==Real)
   {
    spool=OpenWebpageSpoolFile(1,url);

    if(spool!=-1)
      {close(spool); mode=Spool;}
   }

 /* Set up the file descriptor for the spool file. */

 if(mode==Real || mode==Fetch || mode==Get)
   {
    spool=OpenWebpageSpoolFile(0,url);

    if(spool==-1)
      {
       if(mode==Real)
          ServerErrorMessage(client,"Cannot open the spooled web page to write");

       PrintMessage(Fatal,"Cannot open the spooled web page to write.");
      }
   }
 else if(mode==Spool)
   {
    spool=OpenWebpageSpoolFile(1,url);

    if(spool==-1)
       mode=SpoolGet;
   }
 if(mode==SpoolGet)
   {
    spool=OpenWebpageSpoolFile(0,url);

    if(spool==-1)
      {
       ServerErrorMessage(client,"Cannot open the spooled web page to write");

       PrintMessage(Fatal,"Cannot open a spooled web page to write.");
      }
   }

 if(mode==SpoolGet || mode==Get)
   {
    outgoing=OpenOutgoingSpoolFile(0);

    if(outgoing==-1)
      {
       if(mode==SpoolGet)
          ServerErrorMessage(client,"Cannot open the spooled outgoing request to write");

       PrintMessage(Fatal,"Cannot open the spooled outgoing request to write.");
      }
   }

 /* Set up the file descriptor for the server host. */

 if(mode==Real || mode==RealNoCache || mode==Fetch)
   {
    char *colon;

    if(argc==4)
      {
       host=(char*)malloc(strlen(argv[3])+1);
       strcpy(host,argv[3]);
      }
    else
       MakeRequestNonProxy(url,path,args,&buffer);

    if((colon=strchr(host,':')))
      {
       *colon++=0;
       if(*colon)
          port=atoi(colon);
      }

    server=OpenClientSocket(host,port);

    if(server==-1)
      {
       if(mode==Real || mode==Fetch)
          ServerFetchFailed(spool,url,host,port,1);
       if(mode==Real || mode==RealNoCache)
          ServerFetchFailed(client,url,host,port,1);

       PrintMessage(Fatal,"Cannot open a socket to the server %s port %d.",host,port);
      }
   }

 /* Do whatever is needed. */

 if(strlen(buffer)<257)
    buffer=(char*)realloc(buffer,257);

 if(mode==Real || mode==RealNoCache)
   {
    do
      {
       write(server,buffer,strlen(buffer));
      }
    while(buffer && buffer[1] && buffer[2] && (buffer=fgets_realloc(buffer,client_file)));

    if(!buffer)
       buffer=(char*)malloc(257);

    while((n=ReadOrTimeout(server,buffer,256,SERVER_TIMEOUT))>0)
      {
       if(mode==Real)
          write(spool,buffer,n);
       write(client,buffer,n);
      }

    if(n<0 && mode==Real)
      {
       lseek(spool,0,SEEK_SET);
       ServerFetchFailed(spool,url,host,port,0);
       PrintMessage(Fatal,"Timed out while reading from server.");
      }
   }
 else if(mode==Fetch)
   {
    do
      {
       write(server,buffer,strlen(buffer));
      }
    while(buffer && buffer[1] && buffer[2] && (buffer=fgets_realloc(buffer,outgoing_file)));

    if(!buffer)
       buffer=(char*)malloc(strlen(url)>256?strlen(url)+32:257);

    sprintf(buffer,"Fetching %s\n",url);
    write(client,buffer,strlen(buffer));

    while((n=ReadOrTimeout(server,buffer,256,SERVER_TIMEOUT))>0)
       write(spool,buffer,n);

    if(n<0)
      {
       lseek(spool,0,SEEK_SET);
       ServerFetchFailed(spool,url,host,port,0);
       PrintMessage(Fatal,"Timed out while reading from server.");
      }
   }
 else if(mode==Spool)
   {
    char *head="HTTP/1.0 400 WWWOFFLE Fetch Failed\r\n";

    n=read(spool,buffer,256);

    if(n==0 || !strncmp(buffer,head,strlen(head)))
       OpenWebpageSpoolFile(-1,url);

    if(n)
       do
         {
          write(client,buffer,n);
         }
       while((n=read(spool,buffer,256)));
   }
 else if(mode==SpoolGet || mode==Get)
   {
    do
      {
       write(outgoing,buffer,strlen(buffer));
      }
    while(buffer && buffer[1] && buffer[2] && (buffer=fgets_realloc(buffer,client_file)));

    if(mode==SpoolGet)
       WillGet(client,url,0);
    WillGet(spool,url,1);
   }
 else /* mode==Index */
   {
    GenerateIndex(client,url);
   }

 /* Close down and exit. */

 if(client>0)
    fclose(client_file);

 if(server>0)
    CloseSocket(server);

 if(outgoing>0)
   {
    if(mode==Fetch)
       fclose(outgoing_file);
    else
       close(outgoing);
   }

 if(spool>0)
    close(spool);

 if(buffer)
    free(buffer);

 return(0);
}


/*++++++++++++++++++++++++++++++++++++++
  Print the program usage.
  ++++++++++++++++++++++++++++++++++++++*/

static void usage(void)
{
 fputs("\n"
       "WWWOFFLES - World Wide Web Offline Explorer (Server) - Version 0.9\n"
       "\n",stderr);

 fputs("(c) Andrew M. Bishop 1996,97 [       amb@gedanken.demon.co.uk ]\n"
       "                             [http://www.gedanken.demon.co.uk/]\n"
       "\n",stderr);

 fputs("*** Don't run this program, use wwwoffle or wwwoffled instead. ***\n"
       "\n",stderr);

 exit(1);
}
